"""
Entities
"""
from ruruki import interfaces


class Entity(interfaces.IEntity):
    """
    Base class for containing the common methods used for the other
    entities like vertices and edges.

    .. note::

        See :class:`~.IEntity` for doco.

    :param label: :class:`.IEntity` label.
    :type ident: :class:`str` or :obj:`None`
    :param kwargs: Additional properties for the :class:`.IEntity`.
    :type kwargs: :class:`str`=value or :class:`dict`
    """
    __slots__ = ["ident", "label", "properties", "graph"]

    def __init__(self, label=None, **kwargs):
        self.graph = None
        self.ident = None
        self.label = label
        self.properties = kwargs

    def is_bound(self):
        return self.graph is not None

    def remove_property(self, key):
        if key in self.properties:
            del self.properties[key]

    def _update_properties(self, kwargs):
        self.properties.update(kwargs)

    def set_property(self, **kwargs):
        if not kwargs:
            raise interfaces.EntityUpdateError(
                "Can not update with no key and values."
            )

        if self.is_bound():
            self.graph.set_property(self, **kwargs)
        self._update_properties(kwargs)

    def as_dict(self):
        return {
            "metadata": {
            },
            "id": self.ident,
            "label": self.label,
            "properties": self.properties,
        }

    def __str__(self):
        return "<{0}> {1}".format(
            self.__class__.__name__, self.ident
        )

    def __repr__(self):  # pragma: no cover
        return "<{0}> ident: {1}, label: {2}, properties: {3}".format(
            self.__class__.__name__, self.ident, self.label, self.properties
        )


class Vertex(interfaces.IVertex, Entity):
    """
    Vertex/Node is the representation of a entity. It can be anything
    and contains properties for additional information.

    .. note::

        See :class:`~.IVertex` for doco.

    :param label: :class:`.IEntity` label.
    :type ident: :class:`str` or :obj:`None`
    :param kwargs: Additional properties for the :class:`.IEntity`.
    :type kwargs: :class:`str`=value or :class:`dict`
    """
    __slots__ = ["in_edges", "out_edges"]

    def __init__(self, label=None, **kwargs):
        super(Vertex, self).__init__(label=label, **kwargs)
        self.in_edges = EntitySet()
        self.out_edges = EntitySet()

    def in_edge_count(self):
        return len(self.in_edges)

    def out_edge_count(self):
        return len(self.out_edges)

    def add_in_edge(self, vertex, label=None, **kwargs):
        # if the vertex is bound to a graph, then let the graph
        # handle the edge creation.
        if self.is_bound():
            return self.graph.add_edge(vertex, label, self, **kwargs)

        edge = Edge(vertex, label, self, **kwargs)
        self.in_edges.add(edge)
        return edge

    def add_out_edge(self, vertex, label, **kwargs):
        # if the vertex is bound to a graph, then let the graph
        # handle the edge creation.
        if self.is_bound():
            return self.graph.add_edge(self, label, vertex, **kwargs)

        edge = Edge(self, label, vertex, **kwargs)
        self.out_edges.add(edge)
        return edge

    def remove_edge(self, edge):
        head = edge.head
        tail = edge.tail
        if head == self:
            self.out_edges.remove(edge)
        elif tail == self:
            self.in_edges.remove(edge)
        else:
            raise interfaces.VertexError(
                "Unknown edge to this vertex: {}".format(edge)
            )

    def get_in_edges(self, label=None, **kwargs):
        return self.in_edges.filter(label, **kwargs)

    def get_out_edges(self, label=None, **kwargs):
        return self.out_edges.filter(label, **kwargs)

    def get_both_edges(self, label=None, **kwargs):
        edges = self.in_edges | self.out_edges
        return edges.filter(label, **kwargs)  # pylint: disable=no-member

    def get_in_vertices(self, label=None, **kwargs):
        vertices = [
            each.get_in_vertex() for each in self.get_in_edges()
        ]
        return EntitySet(vertices).filter(label, **kwargs)

    def get_out_vertices(self, label=None, **kwargs):
        vertices = [
            each.get_out_vertex() for each in self.get_out_edges()
        ]
        return EntitySet(vertices).filter(label, **kwargs)

    def get_both_vertices(self, label=None, **kwargs):
        in_set = self.get_in_vertices(label=label, **kwargs)
        out_set = self.get_out_vertices(label=label, **kwargs)
        return in_set | out_set

    def as_dict(self):
        return {
            "metadata": {
                "in_edge_count": self.in_edge_count(),
                "out_edge_count": self.out_edge_count(),
            },
            "id": self.ident,
            "label": self.label,
            "properties": self.properties,
        }


class PersistentVertex(Vertex):
    """
    Persistent Vertex behaves exactly the same as a :class:`~.Vertex` but has
    an additional path attribute which is the disk location.
    """
    __slots__ = ["path"]

    def __init__(self, *args, **kwargs):
        super(PersistentVertex, self).__init__(*args, **kwargs)
        self.path = None


class Edge(interfaces.IEdge, Entity):
    """
    Edge/Relationship is the representation of a relationship between two
    entities. A edge has properties for additional information.

    .. note::

        See :class:`~.IEdge` for doco.

    :param head: Head :class:`.IVertex` of the edge.
    :type head: :class:`.IVertex`
    :param label: :class:`.IEntity` label.
    :type ident: :class:`str` or :obj:`None`
    :param tail: Tail :class:`.IVertex` of the edge.
    :type tail: :class:`.IVertex`
    :param kwargs: Additional properties for the :class:`.IEntity`.
    :type kwargs: :class:`str`=value or :class:`dict`
    """
    __slots__ = ["head", "tail"]

    def __init__(self, head, label, tail, **kwargs):
        super(Edge, self).__init__(label=label, **kwargs)
        self.head = head
        self.tail = tail

    def get_in_vertex(self):
        return self.head

    def get_out_vertex(self):
        return self.tail

    def as_dict(self):
        return {
            "metadata": {},
            "id": self.ident,
            "label": self.label,
            "head_id": self.head.ident,
            "tail_id": self.tail.ident,
            "properties": self.properties,
        }

    def __str__(self):  # pragma: no cover
        return "<{0}> ident: {1} [{3}-{2}-{4}]".format(
            self.__class__.__name__, self.ident, self.label,
            self.head.ident, self.tail.ident
        )

    def __repr__(self):  # pragma: no cover
        return (
            "<{0}> ident: {1}, label: {2}, properties: "
            "{3} [{4}-{2}-{5}]".format(
                self.__class__.__name__, self.ident, self.label,
                self.properties, self.head.ident, self.tail.ident
            )
        )


class PersistentEdge(Edge):
    """
    Persistent Edge behaves exactly the same as a :class:`~.Edge` but has an
    additional path attribute which is the disk location.
    """
    __slots__ = ["path"]

    def __init__(self, *args, **kwargs):
        super(PersistentEdge, self).__init__(*args, **kwargs)
        self.path = None


def _split_key_into_noun_verb(key):
    """
    Internal helper function that takes the key and splits it into the
    noun and verb, and returns the noun and verb.

    .. note::

        Example of a key with the special operator.

        key: name__contains
        return: name, contains

    :param key: Key that you are splitting into the noun and verb. The key
        should end with __<operator>
    :type key: :class:`str`
    :returns: Key name and the operator.
    :rtype: :class:`tuple` (:class:`str`, :class:`str` or :obj:`None`)
    """
    split = key.rsplit("__", 1)
    if len(split) == 2:
        return split[0], split[1]
    return key, None


def _contains(prop_value, cmp_value, ignore_case=False):
    """
    Helper function that take two arguments and checks if :param cmp_value:
    is in :param prop_value:.

    :param prop_value: Property value that you are checking.
    :type prop_value: :class:`str`
    :param cmp_value: Value that you are checking if it is in the property
        value.
    :type cmp_value: :class:`str`
    :param ignore_case: True to run using incase sensitive.
    :type ignore_case: :class:`bool`
    :returns: True if :param cmp_value: is in :param prop_value:
    :rtype: class:`bool`
    """
    if ignore_case is True:
        prop_value = prop_value.lower()
        cmp_value = cmp_value.lower()
    return cmp_value in prop_value


def _startswith(prop_value, cmp_value, ignore_case=False):
    """
    Helper function that take two arguments and checks if :param prop_value:
    startswith :param cmp_value:

    :param prop_value: Property value that you are checking.
    :type prop_value: :class:`str`
    :param cmp_value: Value that you are checking if it is in the property
        value startswith.
    :type cmp_value: :class:`str`
    :param ignore_case: True to run using incase sensitive.
    :type ignore_case: :class:`bool`
    :returns: True if :param prop_value: startswith :param cmp_value:
    :rtype: class:`bool`
    """
    if ignore_case is True:
        prop_value = prop_value.lower()
        cmp_value = cmp_value.lower()
    return prop_value.startswith(cmp_value)


def _endswith(prop_value, cmp_value, ignore_case=False):
    """
    Helper function that take two arguments and checks if :param prop_value:
    endswith :param cmp_value:

    :param prop_value: Property value that you are checking.
    :type prop_value: :class:`str`
    :param cmp_value: Value that you are checking if it is in the property
        value endswith.
    :type cmp_value: :class:`str`
    :param ignore_case: True to run using incase sensitive.
    :type ignore_case: :class:`bool`
    :returns: True if :param prop_value: endswith :param cmp_value:
    :rtype: class:`bool`
    """
    if ignore_case is True:
        prop_value = prop_value.lower()
        cmp_value = cmp_value.lower()
    return prop_value.endswith(cmp_value)


def _eq(prop_value, cmp_value, ignore_case=False):
    """
    Helper function that take two arguments and checks if :param prop_value:
    equals :param cmp_value:

    :param prop_value: Property value that you are checking.
    :type prop_value: :class:`str`
    :param cmp_value: Value that you are checking if they are equal.
    :type cmp_value: :class:`str`
    :param ignore_case: True to run using incase sensitive.
    :type ignore_case: :class:`bool`
    :returns: True if :param prop_value: and :param cmp_value: are
        equal.
    :rtype: class:`bool`
    """
    if ignore_case is True:
        prop_value = prop_value.lower()
        cmp_value = cmp_value.lower()
    return cmp_value == prop_value


def _ne(prop_value, cmp_value, ignore_case=False):
    """
    Helper function that take two arguments and checks if :param prop_value:
    is not equal to :param cmp_value:

    :param prop_value: Property value that you are checking.
    :type prop_value: :class:`str`
    :param cmp_value: Value that you are checking if they are not equal.
    :type cmp_value: :class:`str`
    :param ignore_case: True to run using incase sensitive.
    :type ignore_case: :class:`bool`
    :returns: True if :param prop_value: and :param cmp_value: are
        not equal.
    :rtype: class:`bool`
    """
    if ignore_case is True:
        prop_value = prop_value.lower()
        cmp_value = cmp_value.lower()
    return cmp_value != prop_value


OPERATORS = {
    "contains": _contains,
    "icontains": _contains,  # require to be called with ignore_case
    "startswith": _startswith,
    "istartswith": _startswith,  # require to be called with ignore_case
    "endswith": _endswith,
    "iendswith": _endswith,  # require to be called with ignore_case
    "le": lambda prop_value, value, ignore_case: value >= prop_value,
    "lt": lambda prop_value, value, ignore_case: value > prop_value,
    "ge": lambda prop_value, value, ignore_case: value <= prop_value,
    "gt": lambda prop_value, value, ignore_case: value < prop_value,
    "eq": _eq,
    "ieq": _eq,  # require to be called with ignore_case
    "ne": _ne,
    "ine": _ne,  # require to be called with ignore_case
}


class EntitySet(interfaces.IEntitySet):
    """
    EntitySet used for storing, filtering, and iterating over
    :class:`~.IEntity` objects.

    .. note::

        See :class:`~.IEntitySet` for documenation.

    :param entities: Entities being added to the set.
    :type entities: Iterable of :class:`.IEntity`
    """
    def __init__(self, entities=None):
        super(EntitySet, self).__init__()
        self._prop_reference = {}
        self._id_reference = {}

        if entities is not None:
            for entity in entities:
                self.add(entity)

    def all(self, label=None, **kwargs):
        return list(self.filter(label, **kwargs))

    def sorted(self, key=None, reverse=False):
        return sorted(self, key=key, reverse=reverse)

    def get_labels(self):
        return self._prop_reference.keys()

    def get_indexes(self):
        for label in self._prop_reference.iterkeys():
            for key in self._prop_reference[label].iterkeys():
                if not key.startswith("_all"):
                    yield label, key

    def get(self, ident):
        entity = self._id_reference.get(ident)
        if entity is None:
            raise KeyError("No such id {0!r} exists.".format(ident))
        return entity

    def update_index(self, entity, **kwargs):
        collection = self._prop_reference.setdefault(
            entity.label,
            {"_all": set()},
        )

        collection["_all"].add(entity)
        # Add in a indexed property reference.
        for key in kwargs:
            collection.setdefault(key, set()).add(entity)

    def add(self, entity):
        if entity.ident in self._id_reference:
            if entity != self._id_reference[entity.ident]:
                raise KeyError(
                    "Conflict: {0} (current) <-> {1} (conflict)".format(
                        self._id_reference[entity.ident], entity
                    )
                )

        # Add in a reference for fast id search.
        self._id_reference[entity.ident] = entity
        self.update_index(entity, **entity.properties)

        super(EntitySet, self).add(entity)

    def remove(self, entity):
        if entity.ident in self._id_reference:
            del self._id_reference[entity.ident]
        else:
            raise KeyError("No such id {0!r} exists.".format(entity.ident))

        collection = self._prop_reference[entity.label]
        for key in entity.properties.iterkeys():
            if key in collection:
                collection[key].discard(entity)

        super(EntitySet, self).remove(entity)

    def filter(self, label=None, **kwargs):  # pylint: disable=too-many-locals,too-many-branches
        if label is None and not kwargs:
            return self

        if label and not kwargs:
            if label in self._prop_reference:
                return EntitySet(entities=self._prop_reference[label]["_all"])

        keys_values = kwargs.items()
        get_func = OPERATORS.get
        noun_verb_cache = {
            key: _split_key_into_noun_verb(key)
            for key, value in keys_values
        }

        elements = set()
        if label is None:
            elements = set(self._id_reference.values())
        elif label in self._prop_reference:
            for key, value in keys_values:
                key, verb = noun_verb_cache[key]
                if key not in self._prop_reference[label]:
                    return EntitySet()
                elements = elements | self._prop_reference[label][key]

        container = EntitySet()
        for entity in elements:
            mismatch = False
            for key, value in keys_values:
                key, verb = noun_verb_cache[key]
                icase = verb[0] == "i" if verb else False
                func = get_func(verb)
                if key not in entity.properties:
                    mismatch = True
                    break
                prop_value = entity.properties[key]
                if prop_value is None:
                    mismatch = True
                    break
                if not func:
                    if prop_value != value:
                        mismatch = True
                        break
                elif not func(prop_value, value, icase):
                    mismatch = True
                    break

            if not mismatch:
                container.add(entity)

        return container



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods

import unittest2
from ruruki.interfaces import IEdge, VertexError, EntityUpdateError
from ruruki.entities import Entity, Edge, Vertex
from ruruki.test_utils import base
from ruruki.graphs import Graph


class TestEntityBase(unittest2.TestCase):
    def setUp(self):
        self.marko = Entity(label="person", name="marko", age=29)

    def test_is_bound(self):
        graph = Graph()
        self.marko.graph = graph  # mock is bound
        self.assertEqual(
            self.marko.is_bound(),
            True
        )

    def test_is_bound_not_bound(self):
        self.assertEqual(
            self.marko.is_bound(),
            False
        )

    def test_remove_property(self):
        self.marko.remove_property("name")
        self.assertDictEqual(
            self.marko.properties,
            {
                "age": 29,
            }
        )

    def test_remove_property_no_such_key(self):
        self.marko.remove_property("surname")
        self.assertDictEqual(
            self.marko.properties,
            {
                "name": "marko",
                "age": 29,
            }
        )

    def test_set_property_kwargs(self):
        self.marko.set_property(surname="McDonald")
        self.assertDictEqual(
            self.marko.properties,
            {
                "name": "marko",
                "surname": "McDonald",
                "age": 29,
            }
        )

    def test_set_property_no_kwargs(self):
        self.assertRaises(
            EntityUpdateError,
            self.marko.set_property,
        )

    def test_set_property_args_and_kwargs(self):
        self.marko.set_property(surname="McDonald", age=30)
        self.assertDictEqual(
            self.marko.properties,
            {
                "name": "marko",
                "surname": "McDonald",
                "age": 30,
            }
        )

    def test_as_dict(self):
        self.assertDictEqual(
            self.marko.as_dict(),
            {
                "id": None,
                "label": "person",
                "properties": {
                    "name": "marko",
                    "age": 29
                },
                "metadata": {},
            },
        )


class TestVertex(base.TestBase, TestEntityBase):
    def setUp(self):
        self.marko = Vertex(label="person", name="marko", age=29)
        super(TestVertex, self).setUp()

    def test_is_bound_not_bound(self):
        self.marko.graph = None
        self.assertEqual(
            self.marko.is_bound(),
            False
        )

    def test_add_in_edge_bound_to_graph(self):
        sue = self.graph.get_or_create_vertex("person", name="sue")
        new_edge = sue.add_in_edge(self.josh, "brother")

        self.assertIsInstance(
            new_edge,
            IEdge
        )

        self.assertIn(
            new_edge,
            sue.in_edges,
        )

        self.assertEqual(
            sue.get_both_edges().sorted(),
            sorted(
                [
                    new_edge
                ]
            )
        )

    def test_add_in_edge_unbound(self):
        marko = Vertex("person", name="marko")
        josh = Vertex("person", name="josh")
        new_edge = marko.add_in_edge(josh, "brother")

        self.assertIsInstance(
            new_edge,
            IEdge
        )

        self.assertIn(
            new_edge,
            marko.in_edges,
        )

        self.assertEqual(
            marko.get_both_edges().sorted(),
            [new_edge],
        )

    def test_add_out_edge_unbound(self):
        marko = Vertex("person", name="marko")
        josh = Vertex("person", name="josh")
        new_edge = marko.add_out_edge(josh, "knows")

        self.assertIsInstance(
            new_edge,
            IEdge
        )

        self.assertIn(
            new_edge,
            marko.out_edges,
        )

        self.assertEqual(
            marko.get_both_edges().sorted(),
            [new_edge],
        )

    def test_add_out_edge(self):
        sue = self.graph.get_or_create_vertex("person", name="sue")
        new_edge = sue.add_out_edge(self.josh, "sister")

        self.assertIsInstance(
            new_edge,
            IEdge
        )

        self.assertIn(
            new_edge,
            sue.out_edges,
        )

        self.assertEqual(
            sue.get_both_edges().sorted(),
            sorted([new_edge]),
        )

    def test_in_edge_count(self):
        self.marko.add_in_edge(self.josh, "brother")
        self.assertEqual(
            self.marko.in_edge_count(),
            len(self.marko.get_in_edges()),
        )

    def test_out_edge_count(self):
        self.assertEqual(
            self.marko.out_edge_count(),
            len(self.marko.get_out_edges()),
        )

    def test_out_edge_count_remove_edge(self):
        edge = self.marko.get_out_edges().all()[0]
        self.marko.remove_edge(edge)
        self.assertEqual(
            self.marko.out_edge_count(),
            len(self.marko.get_out_edges()),
        )

    def test_in_edge_count_remove_edge(self):
        edge = self.josh.get_in_edges().all()[0]
        self.josh.remove_edge(edge)
        self.assertEqual(
            self.marko.in_edge_count(),
            len(self.marko.get_in_edges()),
        )

    def test_remove_edge(self):
        self.marko.remove_edge(self.marko_created_lop)
        self.assertEqual(
            self.marko.get_both_edges().sorted(),
            sorted([self.marko_knows_josh, self.marko_knows_vadas])
        )

    def test_remove_edge_with_unknown_head(self):
        self.assertRaises(
            VertexError,
            self.marko.remove_edge,
            self.josh_created_lop
        )

    def test_get_in_edges(self):
        self.assertEqual(
            self.josh.get_in_edges().sorted(),
            sorted([self.marko_knows_josh])
        )

    def test_get_in_edges_with_property(self):
        self.assertEqual(
            self.josh.get_in_edges(weight=1).sorted(),
            sorted([self.marko_knows_josh])
        )

    def test_get_in_edges_with_property_that_contains(self):
        self.marko_knows_josh.set_property(since="work")
        self.assertEqual(
            self.josh.get_in_edges(since__contains="r").sorted(),
            sorted([self.marko_knows_josh])
        )

    def test_get_in_edges_with_property_that_startswith(self):
        self.marko_knows_josh.set_property(since="work")
        self.assertEqual(
            self.josh.get_in_edges(since__startswith="w").sorted(),
            sorted([self.marko_knows_josh])
        )

    def test_get_in_edges_with_property_that_endswith(self):
        self.marko_knows_josh.set_property(since="work")
        self.assertEqual(
            self.josh.get_in_edges(since__endswith="k").sorted(),
            sorted([self.marko_knows_josh])
        )

    def test_get_out_edges(self):
        self.assertEqual(
            self.josh.get_out_edges().sorted(),
            sorted([self.josh_created_lop, self.josh_created_ripple])
        )

    def test_get_out_edges_with_property(self):
        self.assertEqual(
            self.josh.get_out_edges(weight=1).sorted(),
            sorted([self.josh_created_ripple])
        )

    def test_get_out_edges_with_property_that_contains(self):
        self.josh_created_lop.set_property(since="work")
        self.josh_created_ripple.set_property(since="class")
        self.assertEqual(
            self.josh.get_out_edges(since__contains="o").sorted(),
            sorted([self.josh_created_lop])
        )

    def test_get_out_edges_with_property_that_startswith(self):
        self.josh_created_lop.set_property(since="work")
        self.josh_created_ripple.set_property(since="class")
        self.assertEqual(
            self.josh.get_out_edges(since__startswith="w").sorted(),
            sorted([self.josh_created_lop])
        )

    def test_get_out_edges_with_property_that_endswith(self):
        self.josh_created_lop.set_property(since="work")
        self.josh_created_ripple.set_property(since="class")
        self.assertEqual(
            self.josh.get_out_edges(since__endswith="s").sorted(),
            sorted([self.josh_created_ripple])
        )

    def test_get_both_edges(self):
        self.assertEqual(
            self.josh.get_both_edges().sorted(),
            sorted(
                [
                    self.josh_created_lop,
                    self.josh_created_ripple,
                    self.marko_knows_josh,
                ]
            )
        )

    def test_get_both_edges_with_property(self):
        self.assertEqual(
            self.josh.get_out_edges().sorted(),
            sorted(
                [
                    self.josh_created_lop,
                    self.josh_created_ripple,
                ]
            )
        )

    def test_get_both_edges_with_property_that_contains(self):
        e = self.graph.get_or_create_edge(
            self.marko,
            "friend",
            self.josh,
            since="school",
        )
        x = self.marko.get_in_edges() | self.marko.get_out_edges()
        x.update_index(e, **e.properties)  # pylint: disable=no-member
        self.assertEqual(
            self.marko.get_both_edges(since__contains="o").all(),
            [e],
        )

    def test_get_both_edges_with_property_that_startswith(self):
        e = self.graph.get_or_create_edge(
            self.marko,
            "friend",
            self.josh,
            since="school",
        )
        self.assertEqual(
            self.marko.get_both_edges(since__startswith="s").all(),
            [e],
        )

    def test_get_both_edges_with_property_that_endswith(self):
        e = self.graph.get_or_create_edge(
            self.marko,
            "friend",
            self.josh,
            since="school",
        )
        self.assertEqual(
            self.marko.get_both_edges(since__endswith="l").all(),
            [e],
        )

    def test_get_in_vertices(self):
        self.assertEqual(
            self.josh.get_in_vertices().sorted(),
            sorted([self.marko])
        )

    def test_get_in_vertices_by_label(self):
        self.assertEqual(
            self.josh.get_in_vertices(label="person").sorted(),
            sorted([self.marko])
        )

    def test_get_in_vertices_by_property(self):
        self.assertEqual(
            self.josh.get_in_vertices(name="marko").sorted(),
            sorted([self.marko])
        )

    def test_get_in_vertices_by_property_that_contains(self):
        self.assertEqual(
            self.josh.get_in_vertices(name__contains="r").sorted(),
            sorted([self.marko])
        )

    def test_get_in_vertices_by_property_that_startswith(self):
        self.assertEqual(
            self.josh.get_in_vertices(name__startswith="m").sorted(),
            sorted([self.marko])
        )

    def test_get_in_vertices_by_property_that_endswith(self):
        self.assertEqual(
            self.josh.get_in_vertices(name__endswith="o").sorted(),
            sorted([self.marko])
        )

    def test_get_out_vertices(self):
        self.assertEqual(
            self.josh.get_out_vertices().sorted(),
            sorted([self.lop, self.ripple])
        )

    def test_get_out_vertices_by_label(self):
        self.assertEqual(
            self.josh.get_out_vertices(label="app").sorted(),
            sorted([self.lop, self.ripple])
        )

    def test_get_out_vertices_by_property(self):
        self.assertEqual(
            self.josh.get_out_vertices(name="ripple").sorted(),
            sorted([self.ripple])
        )

    def test_get_out_vertices_by_property_that_contains(self):
        self.assertEqual(
            self.josh.get_out_vertices(name__contains="i").sorted(),
            sorted([self.ripple])
        )

    def test_get_out_vertices_by_property_that_startswith(self):
        self.assertEqual(
            self.josh.get_out_vertices(name__startswith="l").sorted(),
            sorted([self.lop])
        )

    def test_get_out_vertices_by_property_that_endswith(self):
        self.assertEqual(
            self.josh.get_out_vertices(name__endswith="p").sorted(),
            sorted([self.lop])
        )

    def test_get_both_vertices(self):
        self.assertEqual(
            self.josh.get_both_vertices().sorted(),
            sorted([self.marko, self.lop, self.ripple])
        )

    def test_get_both_vertices_by_label(self):
        self.assertEqual(
            self.josh.get_both_vertices(label="person").sorted(),
            sorted([self.marko])
        )

    def test_get_both_vertices_by_property(self):
        self.assertEqual(
            self.josh.get_both_vertices(age=29).sorted(),
            sorted([self.marko])
        )

    def test_get_both_vertices_by_property_that_contains(self):
        self.assertEqual(
            self.josh.get_both_vertices(name__contains="r").sorted(),
            sorted([self.marko, self.ripple])
        )

    def test_get_both_vertices_by_property_that_startswith(self):
        self.assertEqual(
            self.josh.get_both_vertices(name__startswith="m").sorted(),
            sorted([self.marko])
        )

    def test_get_both_vertices_by_property_that_endsswith(self):
        self.assertEqual(
            self.josh.get_both_vertices(name__endswith="o").sorted(),
            sorted([self.marko])
        )

    def test_as_dict(self):
        self.assertDictEqual(
            self.marko.as_dict(),
            {
                "id": 0,
                "label": "person",
                "metadata": {
                    "in_edge_count": 0,
                    "out_edge_count": 3,
                },
                "properties": {
                    "name": "marko",
                    "age": 29,
                },
            }
        )


class TestEdge(base.TestBase, TestEntityBase):
    def setUp(self):
        self.marko = Vertex(label="person", name="marko", age=29)
        super(TestEdge, self).setUp()

    def test_is_bound(self):
        edge = Edge(self.marko, "de-friends", self.josh)
        graph = Graph()
        graph.bind_to_graph(edge)
        self.assertEqual(
            edge.is_bound(),
            True
        )

    def test_is_bound_not_bound(self):
        edge = Edge(self.marko, "de-friends", self.josh)
        self.assertEqual(
            edge.is_bound(),
            False
        )

    def test_remove_property(self):
        self.marko_knows_josh.remove_property("weight")
        self.assertEqual(self.marko_knows_josh.properties.keys(), [])

    def test_get_out_vertex(self):
        self.assertEqual(self.marko_knows_josh.get_out_vertex(), self.josh)

    def test_get_in_vertex(self):
        self.assertEqual(self.marko_knows_josh.get_in_vertex(), self.marko)

    def test_as_dict(self):
        self.assertDictEqual(
            self.marko_knows_josh.as_dict(),
            {
                "id": 3,
                "label": "knows",
                "metadata": {},
                "properties": {
                    "weight": 1,
                },
                "head_id": 0,
                "tail_id": 3,
            }
        )



#pylint: skip-file
from ruruki.graphs import Graph

# create a empty graph
graph = Graph()


# create some constraints to ensure uniqueness
graph.add_vertex_constraint("person", "name")
graph.add_vertex_constraint("book", "title")
graph.add_vertex_constraint("author", "fullname")
graph.add_vertex_constraint("category", "name")


# add in a couple books and their authors
# Note that I am duplicating the books title property as a name too. This is for ruruki-eye display
programming = graph.get_or_create_vertex("category", name="Programming")
operating_systems = graph.get_or_create_vertex("category", name="Operating Systems")


python_crash_course = graph.get_or_create_vertex("book", title="Python Crash Course")
graph.set_property(python_crash_course, name="Python Crash Course")
eric_matthes = graph.get_or_create_vertex("author", fullname="Eric Matthes", name="Eric", surname="Matthes")
graph.get_or_create_edge(python_crash_course, "CATEGORY", programming)
graph.get_or_create_edge(python_crash_course, "BY", eric_matthes)


python_pocket_ref = graph.get_or_create_vertex("book", title="Python Pocket Reference")
graph.set_property(python_pocket_ref, name="Python Pocket Reference")
mark_lutz = graph.get_or_create_vertex("author", fullname="Mark Lutz", name="Mark", surname="Lutz")
graph.get_or_create_edge(python_pocket_ref, "CATEGORY", programming)
graph.get_or_create_edge(python_pocket_ref, "BY", mark_lutz)


how_linux_works = graph.get_or_create_vertex("book", title="How Linux Works: What Every Superuser Should Know", edition="second")
graph.set_property(how_linux_works, name=how_linux_works.properties["title"])
brian_ward = graph.get_or_create_vertex("author", fullname="Brian Ward", name="Brian", surname="Ward")
graph.get_or_create_edge(how_linux_works, "CATEGORY", operating_systems)
graph.get_or_create_edge(how_linux_works, "BY", brian_ward)


linux_command_line = graph.get_or_create_vertex("book", title="The Linux Command Line: A Complete Introduction", edition="first")
graph.set_property(linux_command_line, name=linux_command_line.properties["title"])
william = graph.get_or_create_vertex("author", fullname="William E. Shotts Jr.", name="William", surname="Shotts")
graph.get_or_create_edge(linux_command_line, "CATEGORY", operating_systems)
graph.get_or_create_edge(linux_command_line, "BY", william)


# let add some customers and book out some books
john = graph.get_or_create_vertex("person", name="John", surname="Doe")
graph.get_or_create_edge(john, "READING", python_crash_course)
graph.get_or_create_edge(john, "INTEREST", programming)

jane = graph.get_or_create_vertex("person", name="Jane", surname="Doe")
graph.get_or_create_edge(jane, "LIKE", operating_systems)
graph.get_or_create_edge(jane, "MARRIED-TO", john)
graph.get_or_create_edge(jane, "READING", linux_command_line)
graph.get_or_create_edge(jane, "READING", python_pocket_ref)


print "People:"
print graph.get_vertices("person").all()

print "\nPython books:"
print graph.get_vertices("book", title__icontains="Python").all()

print "\nVertex with identity number 0:"
print repr(graph.get_vertex(0))

print "\nUnion of books with 'Reference' and 'Crash Course' in the title:"
print (graph.get_vertices("book", name__contains="Reference") | graph.get_vertices("book", title__contains="Crash Course")).all()


print "\nPython books with that do not contain 'Crash Course' in the title:"
print (graph.get_vertices("book", name__contains="Python") - graph.get_vertices("book", title__contains="Crash Course")).all()



# pylint: disable=too-many-instance-attributes
"""
Base Test setup.
"""
import unittest2
from ruruki import graphs
from ruruki.test_utils import helpers


class TestBase(unittest2.TestCase):
    """
    Base test class.
    """
    def setUp(self):
        self.graph = graphs.Graph()
        self.graph.load(helpers.get_test_dump_graph_file_handler())

        # See test_utils/small_people_graph.dump
        # get and set each of the vertices and edges for later use.
        self.marko = self.graph.get_vertices(name="marko").all()[0]
        self.vadas = self.graph.get_vertices(name="vadas").all()[0]
        self.lop = self.graph.get_vertices(name="lop").all()[0]
        self.josh = self.graph.get_vertices(name="josh").all()[0]
        self.ripple = self.graph.get_vertices(name="ripple").all()[0]
        self.peter = self.graph.get_vertices(name="peter").all()[0]

        self.peter_created_lop = self.graph.get_edges(
            self.peter, "created", self.lop
        ).all()[0]

        self.josh_created_lop = self.graph.get_edges(
            self.josh, "created", self.lop
        ).all()[0]

        self.marko_created_lop = self.graph.get_edges(
            self.marko, "created", tail=self.lop
        ).all()[0]

        self.marko_knows_josh = self.graph.get_edges(
            self.marko, "knows", self.josh
        ).all()[0]

        self.marko_knows_vadas = self.graph.get_edges(
            self.marko, "knows", self.vadas
        ).all()[0]

        self.josh_created_ripple = self.graph.get_edges(
            self.josh, "created", self.ripple
        ).all()[0]



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-statements
# pylint: disable=too-many-lines

import json
import os
import shutil
import tempfile
import unittest2
from ruruki import interfaces
from ruruki.graphs import Graph, PersistentGraph
from ruruki.entities import Entity, Edge, Vertex
from ruruki.entities import PersistentVertex, PersistentEdge
from ruruki.test_utils import base, helpers


class TestGraph(base.TestBase):
    def test_dump(self):
        self.maxDiff = None
        tmp_file = helpers.create_tmp_file_handler()
        self.graph.dump(tmp_file)
        tmp_file.seek(0)

        # sorted the edges and vertices before comparing the two dicts
        loaded_temp = json.load(tmp_file)
        for key in loaded_temp:
            loaded_temp[key].sort()

        loaded_dump = json.load(helpers.get_test_dump_graph_file_handler())
        for key in loaded_dump:
            loaded_dump[key].sort()

        self.assertDictEqual(
            loaded_temp,
            loaded_dump,
        )

    def test_load(self):
        graph = Graph()
        fh = helpers.get_test_dump_graph_file_handler()
        graph.load(fh)

        marko = self.graph.get_vertex(0)
        vadas = self.graph.get_vertex(1)
        lop = self.graph.get_vertex(2)
        josh = self.graph.get_vertex(3)
        ripple = self.graph.get_vertex(4)
        peter = self.graph.get_vertex(5)

        peter_created_lop = self.graph.get_edge(0)
        josh_created_lop = self.graph.get_edge(1)
        marko_created_lop = self.graph.get_edge(2)
        marko_knows_josh = self.graph.get_edge(3)
        marko_knows_vadas = self.graph.get_edge(4)
        josh_created_ripple = self.graph.get_edge(5)

        # The containers that the graph uses checks the attributes on
        # vertices and edges and not the `id` from the python built-in
        # function.
        self.assertIn(marko, self.graph)
        self.assertEqual(marko.label, "person")
        self.assertEqual(marko.ident, 0)
        self.assertDictEqual(
            marko.properties,
            {
                "age": 29,
                "name": "marko",
            },
        )

        self.assertIn(vadas, self.graph)
        self.assertEqual(vadas.label, "person")
        self.assertEqual(vadas.ident, 1)
        self.assertDictEqual(
            vadas.properties,
            {
                "age": 27,
                "name": "vadas",
            },
        )

        self.assertIn(lop, self.graph)
        self.assertEqual(lop.label, "app")
        self.assertEqual(lop.ident, 2)
        self.assertDictEqual(
            lop.properties,
            {
                "lang": "java",
                "name": "lop",
            },
        )

        self.assertIn(josh, self.graph)
        self.assertEqual(josh.label, "person")
        self.assertEqual(josh.ident, 3)
        self.assertDictEqual(
            josh.properties,
            {
                "age": 32,
                "name": "josh",
            },
        )

        self.assertIn(ripple, self.graph)
        self.assertEqual(ripple.label, "app")
        self.assertEqual(ripple.ident, 4)
        self.assertDictEqual(
            ripple.properties,
            {
                "lang": "java",
                "name": "ripple",
            },
        )

        self.assertIn(peter, self.graph)
        self.assertEqual(peter.label, "person")
        self.assertEqual(peter.ident, 5)
        self.assertDictEqual(
            peter.properties,
            {
                "age": 35,
                "name": "peter",
            },
        )

        self.assertIn(peter_created_lop, self.graph)
        self.assertEqual(
            peter_created_lop.label, "created"
        )
        self.assertEqual(
            peter_created_lop.head, peter
        )
        self.assertEqual(
            peter_created_lop.tail, lop
        )
        self.assertDictEqual(
            peter_created_lop.properties,
            {
                "weight": 0.2
            }
        )

        self.assertIn(josh_created_lop, self.graph)
        self.assertEqual(
            josh_created_lop.label, "created"
        )
        self.assertEqual(
            josh_created_lop.head, josh
        )
        self.assertEqual(
            josh_created_lop.tail, lop
        )
        self.assertDictEqual(
            josh_created_lop.properties,
            {
                "weight": 0.4
            }
        )

        self.assertIn(marko_created_lop, self.graph)
        self.assertEqual(
            marko_created_lop.label, "created"
        )
        self.assertEqual(
            marko_created_lop.head, marko
        )
        self.assertEqual(
            marko_created_lop.tail, lop
        )
        self.assertDictEqual(
            marko_created_lop.properties,
            {
                "weight": 0.4
            }
        )

        self.assertIn(marko_knows_josh, self.graph)
        self.assertEqual(
            marko_knows_josh.label, "knows"
        )
        self.assertEqual(
            marko_knows_josh.head, marko
        )
        self.assertEqual(
            marko_knows_josh.tail, josh
        )
        self.assertDictEqual(
            marko_knows_josh.properties,
            {
                "weight": 1
            }
        )

        self.assertIn(marko_knows_vadas, self.graph)
        self.assertEqual(
            marko_knows_vadas.label, "knows"
        )
        self.assertEqual(
            marko_knows_vadas.head, marko
        )
        self.assertEqual(
            marko_knows_vadas.tail, vadas
        )
        self.assertDictEqual(
            marko_knows_vadas.properties,
            {
                "weight": 0.5
            }
        )

        self.assertIn(josh_created_ripple, self.graph)
        self.assertEqual(
            josh_created_ripple.label, "created"
        )
        self.assertEqual(
            josh_created_ripple.head, josh
        )
        self.assertEqual(
            josh_created_ripple.tail, ripple
        )
        self.assertDictEqual(
            josh_created_ripple.properties,
            {
                "weight": 1
            }
        )

    def test_add_vertex(self):
        node = self.graph.add_vertex(label="NODE")
        self.assertEqual(node.label, "NODE")
        self.assertIsInstance(node, interfaces.IVertex)
        self.assertIn(node, self.graph)

    def test_add_vertex_set_id(self):
        last = max(each.ident for each in self.graph.get_vertices())
        node = self.graph.add_vertex(label="NODE")
        self.assertEqual(node.ident, last + 1)

    def test_get_vertex(self):
        self.assertEqual(self.graph.get_vertex(1), self.vadas)

    def test_get_verticies(self):
        self.assertEqual(
            self.graph.get_vertices().sorted(),
            sorted(
                [
                    self.marko, self.vadas, self.lop,
                    self.josh, self.peter, self.ripple,
                ]
            ),
        )

    def test_get_verticies_by_label(self):
        self.assertEqual(
            sorted(self.graph.get_vertices(label="app")),
            sorted([self.lop, self.ripple]),
        )

    def test_get_verticies_by_unknown_label(self):
        self.assertEqual(
            sorted(self.graph.get_vertices(label="UNKNOWN")),
            [],
        )

    def test_get_verticies_with_property(self):
        self.assertEqual(
            sorted(self.graph.get_vertices(name="josh")),
            [self.josh],
        )

    def test_get_verticies_with_contains(self):
        self.assertEqual(
            sorted(self.graph.get_vertices(name__contains="s")),
            sorted([self.vadas, self.josh]),
        )

    def test_get_verticies_where_startswith(self):
        self.assertEqual(
            sorted(self.graph.get_vertices(name__startswith="m")),
            sorted([self.marko]),
        )

    def test_get_verticies_where_endswith(self):
        self.assertEqual(
            sorted(self.graph.get_vertices(name__endswith="o")),
            sorted([self.marko]),
        )

    def test_remove_vertex(self):
        # First we need to remove all the edges before we can remove the
        # vertex.
        edges = self.lop.get_both_edges().all()
        for e in edges:
            self.graph.remove_edge(e)
        self.graph.remove_vertex(self.lop)
        self.assertNotIn(self.lop, self.graph)

    def test_remove_vertex_that_has_edges(self):
        self.assertRaises(
            interfaces.VertexBoundByEdges,
            self.graph.remove_vertex,
            self.lop
        )

    def test_add_edge(self):
        new_edge = self.graph.add_edge(self.marko, "FRIENDS", self.peter)
        self.assertEqual(new_edge.label, "FRIENDS")
        self.assertEqual(new_edge.head, self.marko)
        self.assertEqual(new_edge.tail, self.peter)
        self.assertIsInstance(new_edge, interfaces.IEdge)
        self.assertIn(new_edge, self.graph)

    def test_add_edge_check_id_are_set(self):
        last = max(each.ident for each in self.graph.get_edges())
        new_edge = self.graph.add_edge(self.marko, "FRIENDS", self.peter)
        self.assertEqual(new_edge.ident, last + 1)  # +1 since the last edge

    def test_get_edge(self):
        self.assertEqual(self.graph.get_edge(1), self.josh_created_lop)

    def test_get_edges(self):
        self.assertEqual(
            sorted(self.graph.get_edges()),
            sorted(
                [
                    self.peter_created_lop,
                    self.josh_created_lop,
                    self.marko_created_lop,
                    self.marko_knows_josh,
                    self.marko_knows_vadas,
                    self.josh_created_ripple,
                ]
            ),
        )

    def test_get_edges_by_with_head(self):
        self.assertEqual(
            sorted(self.graph.get_edges(head=self.marko)),
            sorted(
                [
                    self.marko_created_lop,
                    self.marko_knows_josh,
                    self.marko_knows_vadas,
                ]
            ),
        )

    def test_get_edges_by_with_tail(self):
        self.assertEqual(
            sorted(self.graph.get_edges(tail=self.lop)),
            sorted(
                [
                    self.marko_created_lop,
                    self.josh_created_lop,
                    self.peter_created_lop,
                ]
            ),
        )

    def test_get_edges_by_label(self):
        self.assertEqual(
            sorted(self.graph.get_edges(label="knows")),
            sorted([self.marko_knows_josh, self.marko_knows_vadas])
        )

    def test_get_edges_by_unknown_label(self):
        self.assertEqual(
            sorted(self.graph.get_edges(label="UNKNOWN")),
            [],
        )

    def test_get_edges_with_property(self):
        self.assertEqual(
            sorted(self.graph.get_edges(weight=0.2)),
            sorted([self.peter_created_lop]),
        )

    def test_get_edges_property_that_contains(self):
        marko_friend_josh = self.graph.get_or_create_edge(
            self.marko, "friend", self.josh,
            since="work"
        )
        self.assertEqual(
            self.graph.get_edges(since__contains="o").sorted(),
            sorted([marko_friend_josh]),
        )

    def test_get_edges_property_that_startswith(self):
        marko_friend_josh = self.graph.get_or_create_edge(
            self.marko, "friend", self.josh,
            since="work"
        )
        self.assertEqual(
            self.graph.get_edges(since__startswith="w").sorted(),
            sorted([marko_friend_josh]),
        )

    def test_get_edges_property_that_endswith(self):
        marko_friend_josh = self.graph.get_or_create_edge(
            self.marko, "friend", self.josh,
            since="work"
        )
        self.assertEqual(
            self.graph.get_edges(since__endswith="k").sorted(),
            sorted([marko_friend_josh]),
        )

    def test_remove_edge(self):
        self.graph.remove_edge(self.marko_created_lop)
        self.assertNotIn(self.marko_created_lop, self.graph)

    def test_contains_unsupported_types(self):
        unsupported = "unsupported object"
        self.assertRaises(
            TypeError,
            self.graph.__contains__,
            unsupported,
        )

    def test_contains_vertex(self):
        self.assertEqual(self.graph.__contains__(self.marko), True)

    def test_contains_vertex_not_found(self):
        v = Vertex("person", name="not_in_graph")
        self.assertEqual(self.graph.__contains__(v), False)

    def test_contains_edge(self):
        self.assertEqual(self.graph.__contains__(self.marko_knows_josh), True)

    def test_contains_edge_not_found(self):
        e = Edge(self.marko, "friend", self.josh)
        self.assertEqual(self.graph.__contains__(e), False)

    def test_set_property(self):
        self.graph.set_property(self.marko, new_prop="prop_value")

        self.assertDictEqual(
            self.marko.properties,
            {
                "name": "marko",
                "new_prop": "prop_value",
                "age": 29,
            }
        )

        self.assertEqual(
            sorted(self.graph.vertices.get_indexes()),
            sorted(
                [
                    ("app", "lang"),
                    ("app", "name"),
                    ("person", "age"),
                    ("person", "name"),
                    ("person", "new_prop"),
                ]
            ),
        )

    def test_set_property_on_an_unknown_entity(self):
        dog = Vertex("dog", name="socks")
        self.assertRaises(
            interfaces.UnknownEntityError,
            self.graph.set_property,
            dog,
            name="spot"
        )

    def test_set_property_with_empty_prop_index(self):
        dog = self.graph.get_or_create_vertex("dog", name="socks")
        self.graph.add_vertex_constraint("dog", "color")
        self.graph.set_property(dog, color="brown")
        self.assertDictEqual(
            dog.properties,
            {
                "name": "socks",
                "color": "brown",
            }
        )

    def test_set_property_indexed_entity_property(self):
        self.graph.get_or_create_vertex("dog", name="socks")
        self.graph.add_vertex_constraint("dog", "color")
        # add in another dog that would be indexed.
        self.graph.get_or_create_vertex("dog", name="patch", color="black")
        spot = self.graph.get_or_create_vertex(
            "dog", name="spot", color="white"
        )
        self.graph.set_property(spot, color="brown")
        self.assertDictEqual(
            spot.properties,
            {
                "name": "spot",
                "color": "brown",
            }
        )

    def test_entity_set_property_constraint_violation(self):
        self.graph.add_vertex_constraint("person", "surname")
        self.graph.get_or_create_vertex("person", name="John", surname="Doe")
        jane = self.graph.get_or_create_vertex("person", name="Jane")
        self.assertRaises(
            interfaces.ConstraintViolation,
            jane.set_property,
            surname="Doe",
        )

    def test_set_property_on_edge(self):
        self.graph.set_property(self.marko_knows_josh, new_prop="prop_value")

        self.assertDictEqual(
            self.marko_knows_josh.properties,
            {
                "new_prop": "prop_value",
                "weight": 1,
            }
        )

    def test_set_property_unknown_type(self):
        some_entity = Entity("SomeEntity")
        self.assertRaises(
            TypeError,
            self.graph.set_property,
            some_entity,
            name="should_not_be_set"
        )

        self.assertDictEqual(
            some_entity.properties,
            {},
        )

    def test_get_vertex_constraints(self):
        self.assertEqual(
            sorted(self.graph.get_vertex_constraints()),
            sorted(
                [
                    ("app", "name"),
                    ("person", "name"),
                ]
            ),
        )

    def test_add_vertex_constraint(self):
        current_constraints = self.graph.get_vertex_constraints()
        self.graph.add_vertex_constraint("dog", "name")
        self.assertEqual(
            sorted(self.graph.get_vertex_constraints()),
            sorted(
                [
                    ("app", "name"),
                    ("dog", "name"),
                    ("person", "name"),
                ]
            ),
        )
        self.assertNotIn(
            ("dog", "name"),
            current_constraints,
        )

    def test_bind_to_graph(self):
        sue = Vertex(100, name="Sue")
        self.graph.bind_to_graph(sue)
        self.assertIs(
            sue.graph,
            self.graph
        )

    def test_bind_to_graph_unsupported_type(self):
        sue = Entity(100, name="Sue")
        self.assertRaises(
            interfaces.UnknownEntityError,
            self.graph.bind_to_graph,
            sue
        )


class TestGraphGetOrCreateVertices(base.TestBase):
    def test_add_new(self):
        vertices = self.graph.get_vertices().all()
        v = self.graph.get_or_create_vertex("person", name="mock_vertex")
        self.assertIn(v, self.graph)
        self.assertNotIn(v, vertices)

    def test_add_new_constraint_prop_key_not_in_kwargs(self):
        vertices = self.graph.get_vertices().all()
        self.graph.add_vertex_constraint("person", "surname")
        v = self.graph.get_or_create_vertex("person", name="mock_vertex")
        self.assertIn(v, self.graph)
        self.assertNotIn(v, vertices)

    def test_add_new_constraint_prop_key_in_kwargs(self):
        self.graph.add_vertex_constraint("person", "name")
        self.graph.add_vertex_constraint("person", "surname")
        vertices = self.graph.get_vertices().all()
        v = self.graph.get_or_create_vertex("person", surname="Marko")
        self.assertNotIn(v, vertices)

    def test_multi_add_new_no_constraints(self):
        # clear out the current constraints loaded from the dump
        self.graph._vconstraints = {}
        v = self.graph.get_or_create_vertex("person", name="mock_vertex")
        vertices = self.graph.get_vertices().all()
        dup = self.graph.get_or_create_vertex(
            "person",
            name="mock_vertex",
            age=30,
        )
        self.assertIn(dup, self.graph)
        self.assertNotIn(dup, vertices)
        self.assertNotEqual(dup, v)

    def test_multi_add_new_with_constraints(self):
        # add in a person name constraint
        self.graph.add_vertex_constraint("person", "name")
        v = self.graph.get_or_create_vertex("person", name="mock_vertex")
        vertices = self.graph.get_vertices().all()
        dup = self.graph.get_or_create_vertex(
            "person",
            name="mock_vertex",
            age=30,
        )
        self.assertIn(dup, self.graph)
        self.assertIn(dup, vertices)
        self.assertEqual(dup, v)

    def test_multi_add_new_with_constraints_unique_vertices_by_name(self):
        # add in a person name constraint
        self.graph.add_vertex_constraint("person", "name")
        v = self.graph.get_or_create_vertex(
            "person",
            name="mock_vertex",
        )

        v1 = self.graph.get_or_create_vertex(
            "person",
            name="mock_vertex1",
        )

        self.assertIn(v, self.graph)
        self.assertIn(v1, self.graph)
        self.assertNotEqual(v, v1)

    def test_more_then_one_found_with_no_constraints(self):
        self.graph.add_vertex(
            "person",
            name="mock_vertex",
            age=20
        )

        self.graph.add_vertex(
            "person",
            name="mock_vertex2",
            age=20
        )

        self.assertRaises(
            interfaces.MultipleFoundExpectedOne,
            self.graph.get_or_create_vertex,
            "person", age=20
        )

    def test_no_constraints_filter_found_one_vertex(self):
        graph = Graph()
        graph.add_vertex("dog", name="rex")

        spot = graph.add_vertex("dog", name="spot")
        spot2 = graph.get_or_create_vertex("dog", name="spot")

        self.assertEqual(
            spot2,
            spot,
        )

    def test_no_label_no_kwargs(self):
        self.assertEqual(
            self.graph.get_or_create_vertex(),
            None
        )


class TestGraphGetOrCreateEdges(base.TestBase):
    def test_add_edge_with_head_tuple(self):
        vertices = self.graph.get_vertices("person", name="sue")
        edge = self.graph.get_or_create_edge(
            ("person", {"name": "sue"}),
            "knowns",
            self.marko,
        )

        self.assertIn(
            edge,
            self.graph
        )

        sue = self.graph.get_vertices("person", name="sue").all()[0]
        self.assertNotIn(
            sue,
            vertices,
        )

        self.assertIn(
            sue,
            self.graph,
        )

    def test_add_edge_with_tail_tuple(self):
        vertices = self.graph.get_vertices("person", name="sue")
        edge = self.graph.get_or_create_edge(
            self.marko,
            "knowns",
            ("person", {"name": "sue"}),
        )

        self.assertIn(
            edge,
            self.graph
        )

        sue = self.graph.get_vertices("person", name="sue").all()[0]
        self.assertNotIn(
            sue,
            vertices,
        )

        self.assertIn(
            sue,
            self.graph,
        )

    def test_add_edge_constraint_violation(self):
        self.assertRaises(
            interfaces.ConstraintViolation,
            self.graph.add_edge,
            self.marko,
            "knows",
            self.josh,
            place="work",
        )

    def test_edge_already_indexed(self):
        self.assertEqual(
            self.graph.get_or_create_edge(self.marko, "knows", self.josh),
            self.marko_knows_josh,
        )


def create_graph_mock_path():
    """
    Create the folllowing

    path
       |_ vertices
       |     |_ constraints.json (file)
       |     |_ label
       |     |     |_ 0
       |     |        |_ properties.json (file)
       |     |        |_ in-edges
       |     |        |     |_ 0 -> ../../../../edges/label/0 (symlink)
       |     |        |_ out-edges
       |     |              |_
       |     |
       |     |_ label
       |         |_ 1
       |              |_ properties.json (file)
       |              |_ in-edges
       |              |     |_
       |              |_ out-edges
       |                    |_ 0 -> ../../../../edges/label/0 (symlink)
       |
       |_ edges
             |_ constraints.json
             |_ label
                   |
                   |_0
                     |_ properties.json (file)
                     |_ head
                     |   |_ ../../../vertices/0 (symlik)
                     |_ tail
                         |_ ../../../vertices/1 (symlik)
    """
    path = tempfile.mkdtemp()
    vertices_path = os.path.join(path, "vertices")
    os.makedirs(vertices_path)
    edges_path = os.path.join(path, "edges")
    os.makedirs(edges_path)

    # create constraints
    open(os.path.join(edges_path, "constraints.json"), "w").close()
    json.dump(
        [
            {
                "label": "person",
                "key": "name",
            }
        ],
        open(os.path.join(vertices_path, "constraints.json"), "w"),
        indent=4,
    )

    # create two vertices
    marko_path = os.path.join(vertices_path, "person", "0")
    marko_in_edges_path = os.path.join(marko_path, "in-edges")
    marko_out_edges_path = os.path.join(marko_path, "out-edges")
    os.makedirs(marko_path)
    os.makedirs(marko_in_edges_path)
    os.makedirs(marko_out_edges_path)
    json.dump(
        {"name": "Marko"},
        open(os.path.join(marko_path, "properties.json"), "w"),
        indent=4,
    )

    josh_path = os.path.join(vertices_path, "person", "1")
    josh_in_edges_path = os.path.join(josh_path, "in-edges")
    josh_out_edges_path = os.path.join(josh_path, "out-edges")
    os.makedirs(josh_path)
    os.makedirs(josh_in_edges_path)
    os.makedirs(josh_out_edges_path)
    json.dump(
        {"name": "Josh"},
        open(os.path.join(josh_path, "properties.json"), "w"),
        indent=4,
    )

    # create edge between the vertices
    edge_path = os.path.join(edges_path, "knows", "0")
    head_path = os.path.join(edge_path, "head")
    tail_path = os.path.join(edge_path, "tail")
    os.makedirs(edge_path)
    os.makedirs(head_path)
    os.makedirs(tail_path)
    json.dump(
        {"since": "school"},
        open(os.path.join(edge_path, "properties.json"), "w"),
        indent=4,
    )

    os.symlink(marko_path, os.path.join(head_path, "0"))
    os.symlink(josh_path, os.path.join(tail_path, "1"))

    return path


class TestPersistentGraph(unittest2.TestCase):
    def setUp(self):
        self.graph = PersistentGraph()

    def test_import_from_path(self):
        path = create_graph_mock_path()
        graph = PersistentGraph(path)

        # check that the paths are setup correctly
        self.assertEqual(graph.path, path)

        self.assertEqual(
            graph.vertices_path,
            os.path.join(path, "vertices")
        )
        self.assertEqual(
            graph.vertices_constraints_path,
            os.path.join(path, "vertices", "constraints.json")
        )

        self.assertEqual(
            graph.edges_path,
            os.path.join(path, "edges")
        )

        # check that the constraints have been imported
        self.assertEqual(
            sorted(graph.get_vertex_constraints()),
            [("person", "name")],
        )

    def test_import_from_empty_path_with_auto_create(self):
        path = tempfile.mkdtemp()
        graph = PersistentGraph(path, auto_create=True)

        # check the top level directories have been created.
        self.assertEqual(
            sorted(os.listdir(graph.path)),
            sorted(["vertices", "edges"]),
        )

        # check the vertices directory and other files have been created.
        self.assertEqual(
            sorted(os.listdir(graph.vertices_path)),
            sorted(["constraints.json"]),
        )

    def test_import_from_empty_path_without_auto_create(self):
        path = tempfile.mkdtemp()
        self.assertRaises(
            IOError,
            PersistentGraph,
            path,
            auto_create=False
        )

    def test_import_from_path_auto_create_has_edges_missing_vertices(self):
        path = create_graph_mock_path()
        shutil.rmtree(
            os.path.join(
                path,
                "vertices",
            )
        )

        self.assertRaises(
            OSError,
            PersistentGraph,
            path,
            auto_create=True
        )

    def test_import_from_path_auto_create_has_no_missing_dirs(self):
        path = create_graph_mock_path()
        graph = PersistentGraph(path, auto_create=True)
        self.assertEqual(len(graph.vertices), 2)
        self.assertEqual(len(graph.edges), 1)

    def test_import_from_empty_path_data_written_new_empty_path(self):
        # this is just to check that data is written to the empty
        # path that was provided and automatically setup.
        path = tempfile.mkdtemp()
        graph = PersistentGraph(path, auto_create=True)

        marko = graph.get_or_create_vertex("person", name="Marko")
        spot = graph.get_or_create_vertex("dog", name="Spot")
        graph.get_or_create_edge(marko, "owns", spot)


        # Just check a couple things, because we will assume that if
        # we have some data in the right place, all the reset should be
        # there too.

        self.assertEqual(
            sorted(os.listdir(graph.vertices_path)),
            sorted(["constraints.json", "person", "dog"]),
        )

        self.assertEqual(
            sorted(os.listdir(marko.path)),
            sorted(["in-edges", "out-edges", "properties.json"]),
        )

        self.assertEqual(
            sorted(os.listdir(graph.edges_path)),
            sorted(["owns"]),
        )

    def test_import_with_vertices_missing_properties_file(self):
        path = create_graph_mock_path()
        os.remove(
            os.path.join(
                path,
                "vertices",
                "person",
                "0",
                "properties.json"
            )
        )
        graph = PersistentGraph(path)
        vertex = graph.get_vertex(0)
        self.assertDictEqual(vertex.properties, {})

    def test_import_from_path_missing_vertices_directory(self):
        path = tempfile.mkdtemp()
        os.makedirs(os.path.join(path, "edges"))
        self.assertRaises(
            IOError,
            PersistentGraph,
            path
        )

    def test_import_from_path_missing_vertices_constraints(self):
        path = tempfile.mkdtemp()
        os.makedirs(os.path.join(path, "vertices"))
        os.makedirs(os.path.join(path, "edges"))
        self.assertRaises(
            IOError,
            PersistentGraph,
            path
        )

    def test_import_from_path_missing_edges_directory(self):
        path = tempfile.mkdtemp()
        os.makedirs(os.path.join(path, "vertices"))
        fh = open(os.path.join(path, "vertices", "constraints.json"), "w")
        fh.write("{}")
        fh.close()
        self.assertRaises(
            OSError,
            PersistentGraph,
            path
        )

    def test_import_from_path_loaded_edges(self):
        path = create_graph_mock_path()
        graph = PersistentGraph(path)

        self.assertEqual(len(graph.edges), 1)
        marko_josh = graph.get_edge(0)

        self.assertIsInstance(marko_josh, PersistentEdge)

        self.assertEqual(
            marko_josh.path,
            os.path.join(graph.edges_path, "knows", "0"),
        )

        self.assertDictEqual(
            marko_josh.as_dict(),
            {
                "id": 0,
                "label": "knows",
                "metadata": {},
                "head_id": 0,
                "tail_id": 1,
                "properties": {
                    "since": "school",
                },
            }
        )

    def test_import_from_path_edges_with_bogus_extra_dir(self):
        path = create_graph_mock_path()
        os.mkdir(os.path.join(path, "edges", "knows", "bogus_dir"))
        graph = PersistentGraph(path)
        # should still have loaded the edge
        self.assertEqual(len(graph.edges), 1)

    def test_import_from_path_vertices_with_bogus_extra_dir(self):
        path = create_graph_mock_path()
        os.mkdir(os.path.join(path, "vertices", "person", "bogus_dir"))
        graph = PersistentGraph(path)
        # should still have loaded the edge
        self.assertEqual(len(graph.vertices), 2)

    def test_import_from_path_loaded_edges_missing_properties_file(self):
        path = create_graph_mock_path()

        # deleted the edges properties file.
        os.remove(
            os.path.join(
                path,
                "edges",
                "knows",
                "0",
                "properties.json"
            )
        )

        graph = PersistentGraph(path)

        marko = graph.get_vertex(0)
        josh = graph.get_vertex(1)
        marko_josh = graph.get_edge(0)

        self.assertEqual(len(graph.edges), 1)
        self.assertEqual(marko_josh.head, marko)
        self.assertEqual(marko_josh.tail, josh)
        self.assertIsInstance(marko_josh, PersistentEdge)
        self.assertEqual(
            marko_josh.path,
            os.path.join(graph.edges_path, "knows", "0"),
        )
        self.assertDictEqual(marko_josh.properties, {})

    def test_import_from_path_loaded_edges_unknown_vertex(self):
        path = create_graph_mock_path()
        old = os.path.join(
            path,
            "edges",
            "knows",
            "0",
            "head",
            "0"
        )
        new = os.path.join(
            path,
            "edges",
            "knows",
            "0",
            "head",
            "5"
        )
        os.rename(old, new)

        self.assertRaises(
            KeyError,
            PersistentGraph,
            path
        )

    def test_import_from_path_loaded_edges_head_vertex_not_a_int(self):
        path = create_graph_mock_path()
        old = os.path.join(
            path,
            "edges",
            "knows",
            "0",
            "head",
            "0"
        )
        new = os.path.join(
            path,
            "edges",
            "knows",
            "0",
            "head",
            "non-int"
        )
        os.rename(old, new)

        graph = PersistentGraph(path)
        self.assertEqual(len(graph.edges), 0)

    def test_import_from_path_loaded_edges_tail_vertex_not_a_int(self):
        path = create_graph_mock_path()
        old = os.path.join(
            path,
            "edges",
            "knows",
            "0",
            "tail",
            "1"
        )
        new = os.path.join(
            path,
            "edges",
            "knows",
            "0",
            "tail",
            "non-int"
        )
        os.rename(old, new)

        graph = PersistentGraph(path)
        self.assertEqual(len(graph.edges), 0)

    def test_import_from_path_loaded_vertices(self):
        path = create_graph_mock_path()
        graph = PersistentGraph(path)

        self.assertEqual(len(graph.vertices), 2)

        # test marko was imported and has id 0
        marko = graph.get_vertex(0)
        self.assertIsInstance(marko, PersistentVertex)
        self.assertEqual(
            marko.path,
            os.path.join(graph.vertices_path, "person", "0")
        )
        self.assertDictEqual(
            marko.properties,
            {
                "name": "Marko"
            }
        )

        # test josh was imported and has id 1
        josh = graph.get_vertex(1)
        self.assertIsInstance(josh, PersistentVertex)
        self.assertEqual(
            josh.path,
            os.path.join(graph.vertices_path, "person", "1")
        )
        self.assertDictEqual(
            josh.properties,
            {
                "name": "Josh"
            }
        )

        # check that the next vertex has an id of 3
        spot = graph.add_vertex("dog", name="Spot")
        self.assertEqual(spot.ident, 2)

    def test_create_persistent_graph_with_no_path(self):
        self.assertEqual(
            sorted(os.listdir(self.graph.path)),
            sorted(["edges", "vertices"]),
        )

        # check for the constraints files
        self.assertEqual(
            sorted(os.listdir(self.graph.vertices_path)),
            sorted(["constraints.json"]),
        )

    def test_add_vertex(self):
        marko = self.graph.add_vertex("person", name="Marko")
        josh = self.graph.add_vertex("person", name="Josh")
        spot = self.graph.add_vertex("dog", name="Spot")

        # Check for the label folder
        self.assertEqual(
            sorted(os.listdir(self.graph.vertices_path)),
            sorted(["constraints.json", "dog", "person"]),
        )

        # Check in the label folder for vertex ident folders
        self.assertEqual(
            sorted(
                os.listdir(
                    os.path.join(self.graph.vertices_path, "person")
                )
            ),
            sorted([str(marko.ident), str(josh.ident)]),
        )

        self.assertEqual(
            sorted(
                os.listdir(
                    os.path.join(self.graph.vertices_path, "dog")
                )
            ),
            sorted([str(spot.ident)]),
        )

        # check in the verext folder for the property file and the
        # in/out edges folders - checking only one vertex should be enough.
        self.assertEqual(
            sorted(os.listdir(marko.path)),
            sorted(["properties.json", "in-edges", "out-edges"]),
        )

        # check that the properties file was populated
        self.assertDictEqual(
            json.load(open(os.path.join(str(marko.path), "properties.json"))),
            {"name": "Marko"},
        )

    def test_add_edge(self):
        marko = self.graph.add_vertex("person", name="Marko")
        josh = self.graph.add_vertex("person", name="Josh")
        edge = self.graph.add_edge(marko, "knows", josh, since="school")

        # Check for the label folder
        self.assertEqual(
            sorted(os.listdir(self.graph.edges_path)),
            ["knows"],
        )

        # Check in the label folder for edge ident folders
        self.assertEqual(
            sorted(
                os.listdir(
                    os.path.join(self.graph.edges_path, "knows")
                )
            ),
            sorted([str(edge.ident)]),
        )

        # check in the edge folder has a properties file, and a head and tail
        # directory
        self.assertEqual(
            sorted(
                os.listdir(
                    os.path.join(
                        self.graph.edges_path, "knows", str(edge.ident)
                    )
                )
            ),
            sorted(
                [
                    "properties.json",
                    "head",
                    "tail"
                ]
            ),
        )

        # check that the vertices in the edge folder are symlinks
        mark_link = os.path.join(
            self.graph.edges_path,
            "knows",
            str(edge.ident),
            "head",
            str(marko.ident),
        )

        josh_link = os.path.join(
            self.graph.edges_path,
            "knows",
            str(edge.ident),
            "tail",
            str(josh.ident),
        )

        self.assertEqual(os.path.exists(mark_link), True)
        self.assertEqual(os.path.exists(josh_link), True)
        self.assertEqual(os.path.islink(mark_link), True)
        self.assertEqual(os.path.islink(josh_link), True)

        # check that the properties file was populated
        self.assertEqual(
            json.load(
                open(
                    os.path.join(
                        self.graph.edges_path,
                        "knows",
                        str(edge.ident),
                        "properties.json"
                    )
                )
            ),
            {"since": "school"},
        )

        # check that the edge in the vertex in and out edge folders are
        # symlinked to the edge
        mark_out_edge_link = os.path.join(
            self.graph.vertices_path,
            "person",
            str(marko.ident),
            "out-edges",
            str(edge.ident),
        )

        josh_in_edge_link = os.path.join(
            self.graph.vertices_path,
            "person",
            str(josh.ident),
            "in-edges",
            str(edge.ident),
        )

        self.assertEqual(os.path.islink(mark_out_edge_link), True)
        self.assertEqual(os.path.islink(josh_in_edge_link), True)

    def test_set_property(self):
        marko = self.graph.add_vertex("person", name="Marko")
        self.graph.set_property(marko, surname="Polo")

        self.assertDictEqual(
            json.load(
                open(
                    os.path.join(
                        self.graph.vertices_path,
                        "person",
                        str(marko.ident),
                        "properties.json"
                    )
                )
            ),
            {
                "name": "Marko",
                "surname": "Polo",
            },
        )

    def test_remove_vertex(self):
        marko = self.graph.add_vertex("person", name="Marko")
        josh = self.graph.add_vertex("person", name="Josh")

        self.graph.remove_vertex(marko)

        self.assertEqual(
            sorted(
                os.listdir(
                    os.path.join(self.graph.vertices_path, "person")
                )
            ),
            sorted([str(josh.ident)]),
        )

    def test_remove_edge(self):
        marko = self.graph.add_vertex("person", name="Marko")
        josh = self.graph.add_vertex("person", name="Josh")
        spot = self.graph.add_vertex("dog", name="Spot")

        edge = self.graph.add_edge(marko, "knows", josh)
        edge2 = self.graph.add_edge(marko, "knows", spot)

        self.graph.remove_edge(edge)

        self.assertEqual(
            sorted(
                os.listdir(
                    os.path.join(self.graph.edges_path, "knows")
                )
            ),
            sorted([str(edge2.ident)]),
        )

    def test_add_vertex_constraints(self):
        self.graph.add_vertex_constraint("person", "name")
        print self.graph.vertices_path
        self.assertEqual(
            json.load(
                open(
                    os.path.join(
                        self.graph.vertices_path,
                        "constraints.json"
                    ),
                )
            ),
            [
                {
                    "label": "person",
                    "key": "name"
                }
            ],
        )



"""
Graph implementations
"""
from collections import defaultdict
import json
import logging
import os
import shutil
from tempfile import mkdtemp
from ruruki import interfaces
from ruruki.entities import Vertex, Edge, PersistentVertex, PersistentEdge
from ruruki.entities import EntitySet


def _search_for_edge_ids(path):
    """
    Internal helper function to search for edges identity numbers
    based in the directories and the head and tail vertices identity
    numbers.

    :param path: Edges Path to find number directories.
    :type path: :class:`str`
    :return: Yields the integer number, head id, label, tail id and the
        properties file path if found.
    :rtype: Iterator of :class:`tuple`
        (
            :class:`int`, :class:`int`, :class:`str`,
            :class:`int`, :class:`str` or :obj:`None`
        )
    """
    for label in os.listdir(path):

        label_path = os.path.join(path, label)

        # skip over files because we are only looking for directories.
        if os.path.isfile(label_path):
            continue

        # run over all the edge id's that we can find
        # and loading the properties if found.
        for each in os.listdir(label_path):
            try:
                ident = int(each)
            except ValueError:
                logging.error(
                    "%r is not a expected edge id number, skipping edge import",
                    each
                )
                continue

            propfile = os.path.join(
                label_path,
                each,
                "properties.json"
            )

            try:
                head_id = int(
                    os.listdir(
                        os.path.join(
                            label_path,
                            each,
                            "head"
                        )
                    )[0]
                )
            except ValueError:
                logging.error(
                    "%r is not a expected head id number, skipping edge import",
                    each
                )
                continue

            try:
                tail_id = int(
                    os.listdir(
                        os.path.join(
                            label_path,
                            each,
                            "tail"
                        )
                    )[0]
                )
            except ValueError:
                logging.error(
                    "%r is not a expected tail id number, skipping edge import",
                    each
                )
                continue

            yield (
                ident,
                head_id,
                label,
                tail_id,
                propfile if os.path.isfile(propfile) else None
            )


def _search_for_vertex_id(path):
    """
    Internal helper function to search for vertices identity numbers
    based in the directories.

    :param path: Vertice path to find number directories.
    :type path: :class:`str`
    :return: Yields the integer number, label and the properties file path
        if found.
    :rtype: Iterator of :class:`tuple`
        (:class:`int`, :class:`str`, :class:`str` or :obj:`None`)
    """
    for label in os.listdir(path):
        label_path = os.path.join(path, label)

        # skip over files because we are only looking for directories.
        if os.path.isfile(label_path):
            continue

        # run over all the vertice id's that we can find
        # and loading the properties if found.
        for each in os.listdir(label_path):
            try:
                ident = int(each)
            except ValueError:
                logging.error(
                    "%r is not a expected vertex id number, skipping",
                    each
                )
                continue

            propfile = os.path.join(
                label_path,
                str(ident),
                "properties.json"
            )

            yield ident, label, propfile if os.path.isfile(propfile) else None


class IDGenerator(object):
    """
    ID generator and tracker.
    """

    def __init__(self):
        self.vid = 0
        self.eid = 0

    def get_edge_id(self):
        """
        Generate a edge id.

        :returns: Edge id number.
        :rtype: :class:`int`
        """
        ident = self.eid
        self.eid += 1
        return ident

    def get_vertex_id(self):
        """
        Generate a vertex id.

        :returns: Vertex id number.
        :rtype: :class:`int`
        """
        ident = self.vid
        self.vid += 1
        return ident


class Graph(interfaces.IGraph):
    """
    In-memory graph database.

    See :class:`~.IGraph` for doco.
    """

    def __init__(self):
        self._vclass = Vertex
        self._eclass = Edge
        self._id_tracker = IDGenerator()
        self._vconstraints = defaultdict(dict)
        self._econstraints = defaultdict()
        self.vertices = EntitySet()
        self.edges = EntitySet()

    def load(self, file_handler):
        vertex_id_mapping = {}
        data = json.load(file_handler)

        constraints = data.get("constraints", [])
        for constraint_dict in constraints:
            self.add_vertex_constraint(
                constraint_dict["label"],
                constraint_dict["key"],
            )

        vertices = sorted(data.get("vertices", []), key=lambda x: x["id"])
        for vertex_dict in vertices:
            vertex = self.get_or_create_vertex(
                vertex_dict["label"],
                **vertex_dict["properties"]
            )
            vertex_id_mapping[vertex_dict["id"]] = vertex

        edges = sorted(data.get("edges", []), key=lambda x: x["id"])
        for edge_dict in edges:
            head = vertex_id_mapping[edge_dict["head_id"]]
            tail = vertex_id_mapping[edge_dict["tail_id"]]
            self.get_or_create_edge(
                head,
                edge_dict["label"],
                tail,
                **edge_dict["properties"]
            )

    def dump(self, file_handler):
        data = {
            "vertices": [],
            "edges": [],
            "constraints": [],
        }

        for vertex in self.vertices:
            data["vertices"].append(vertex.as_dict())

        for edge in self.edges:
            data["edges"].append(edge.as_dict())

        for label, key in self.get_vertex_constraints():
            data["constraints"].append(
                {
                    "label": label,
                    "key": key
                }
            )

        json.dump(data, file_handler, indent=4, sort_keys=True)

    def add_vertex_constraint(self, label, key):
        self._vconstraints[label][key] = set()

    def get_vertex_constraints(self):
        constraints = []
        for label in self._vconstraints:
            for key in self._vconstraints[label]:
                constraints.append((label, key))
        return constraints

    def bind_to_graph(self, entity):
        if isinstance(entity, interfaces.IVertex):
            entity.ident = self._id_tracker.get_vertex_id()
        elif isinstance(entity, interfaces.IEdge):
            entity.ident = self._id_tracker.get_edge_id()
        else:
            raise interfaces.UnknownEntityError(
                "Unknown entity {0!r}".format(entity)
            )
        entity.graph = self

    def get_or_create_vertex(self, label=None, **kwargs):
        if not label or not kwargs:
            return None

        # first check constraints.
        if label in self._vconstraints:
            for key, collection in self._vconstraints[label].items():
                if key not in kwargs:
                    continue

                for vertex in collection:
                    if vertex.properties[key] == kwargs[key]:
                        return vertex

        # no matches in constraints, so do a EntitySet filter
        vertices = self.vertices.filter(label, **kwargs).all()
        if len(vertices) > 1:
            raise interfaces.MultipleFoundExpectedOne(
                "Multiple vertices found when one expected."
            )
        elif len(vertices) == 1:
            return vertices[0]

        return self.add_vertex(label, **kwargs)

    def get_or_create_edge(self, head, label, tail, **kwargs):
        if isinstance(head, tuple):
            head = self.get_or_create_vertex(head[0], **head[1])

        if isinstance(tail, tuple):
            tail = self.get_or_create_vertex(tail[0], **tail[1])

        # There can only a single edge between head and tail with a
        # particular label. So there is not point filtering for
        # properties.
        indexed_edge = self._econstraints.get((head, label, tail))
        if indexed_edge:
            return indexed_edge
        return self.add_edge(head, label, tail, **kwargs)

    def add_edge(self, head, label, tail, **kwargs):
        if (head, label, tail) in self._econstraints:
            raise interfaces.ConstraintViolation(
                "Duplicate {0!r} edges between head {1!r} and tail {2!r} "
                "is not allowed".format(
                    label,
                    head,
                    tail
                )
            )
        edge = self._eclass(head, label, tail, **kwargs)
        self._econstraints[(head, label, tail)] = edge
        self.bind_to_graph(edge)
        self.edges.add(edge)
        head.out_edges.add(edge)
        tail.in_edges.add(edge)
        return edge

    def add_vertex(self, label=None, **kwargs):
        vertex = self._vclass(label=label, **kwargs)

        if label in self._vconstraints:
            for key in self._vconstraints[label]:
                if key in kwargs:
                    self._vconstraints[label][key].add(vertex)

        self.bind_to_graph(vertex)
        self.vertices.add(vertex)
        return vertex

    def set_property(self, entity, **kwargs):
        if entity not in self:
            raise interfaces.UnknownEntityError(
                "Unknown entity {0!r}".format(entity)
            )

        if isinstance(entity, interfaces.IVertex):
            key_index = self._vconstraints.get(entity.label, {})
            for key, value in kwargs.items():
                if key not in key_index:
                    continue
                for indexed_entity in key_index[key]:
                    if indexed_entity != entity:
                        if indexed_entity.properties[key] == value:
                            raise interfaces.ConstraintViolation(
                                "Constraint violation with {0}".format(
                                    entity
                                )
                            )
            self.vertices.update_index(entity, **kwargs)

        if isinstance(entity, interfaces.IEdge):
            self.edges.update_index(entity, **kwargs)

        entity._update_properties(kwargs)  # pylint: disable=protected-access

    def get_edge(self, id_num):
        return self.edges.get(id_num)

    def get_vertex(self, id_num):
        return self.vertices.get(id_num)

    def get_edges(self, head=None, label=None, tail=None, **kwargs):
        if head is None and tail is None:
            return self.edges.filter(label, **kwargs)

        container = EntitySet()
        for edge in self.edges.filter(label, **kwargs):
            if head and tail is None:
                if edge.head == head:
                    container.add(edge)
            elif tail and head is None:
                if edge.tail == tail:
                    container.add(edge)
            else:
                if edge.head == head and edge.tail == tail:
                    container.add(edge)
        return container

    def get_vertices(self, label=None, **kwargs):
        return self.vertices.filter(label, **kwargs)

    def remove_edge(self, edge):
        edge.head.remove_edge(edge)
        edge.tail.remove_edge(edge)
        self.edges.remove(edge)

    def remove_vertex(self, vertex):
        if len(vertex.get_both_edges()) > 0:
            raise interfaces.VertexBoundByEdges(
                "Vertex {0!r} is still bound to another vertex "
                "by an edge. First remove all the edges on the vertex and "
                "then remove it again.".format(vertex)
            )
        self.vertices.remove(vertex)

    def close(self):  # pragma: no cover
        # Nothing to do for the close at this stage.
        return

    def __contains__(self, entity):
        is_vertex = isinstance(entity, interfaces.IVertex)
        is_edge = isinstance(entity, interfaces.IEdge)

        if not is_vertex and not is_edge:
            raise TypeError(
                "Unsupported entity type {0}".format(type(entity))
            )

        return entity in self.vertices or entity in self.edges


class PersistentGraph(Graph):
    """
    Persistent Graph database storing data to a file system.

    See :class:`~.IGraph` for doco.

    .. code::

        path
           |_ vertices
           |     |_ constraints.json (file)
           |     |_ label
           |     |     |_ 0
           |     |        |_ properties.json (file)
           |     |        |_ in-edges
           |     |        |     |_ 0 -> ../../../../edges/label/0 (symlink)
           |     |        |_ out-edges
           |     |              |_
           |     |
           |     |_ label
           |     |    |_ 1
           |     |         |_ properties.json (file)
           |     |          |_ in-edges
           |     |          |     |_
           |     |          |_ out-edges
           |     |                |_ 0 -> ../../../../edges/label/0 (symlink)
           |
           |_ edges
                 |_ label
                       |
                       |_0
                         |_ properties.json (file)
                         |_ head
                         |   |_ 0 -> ../../../vertices/0 (symlink)
                         |_ tail
                             |_ 1 -> ../../../vertices/1 (symlink)

    :param path: Path to ruruki graph data on disk. If :obj:`None`, then
        a temporary path will be created, else passing in a path to an
        empty directory will result in the creation of the graph data
        in the provided path.
    :param auto_create: If True, then missing ``vertices`` or ``edges``
        directories will be created.
    :type auto_create: :class:`bool`
    :type path: :class:`str`
    """
    def __init__(self, path=None, auto_create=False):
        super(PersistentGraph, self).__init__()
        self._vclass = PersistentVertex
        self._eclass = PersistentEdge
        self.path = path

        if path is not None:
            self.vertices_path = os.path.join(self.path, "vertices")
            self.edges_path = os.path.join(self.path, "edges")
            self.vertices_constraints_path = os.path.join(
                self.vertices_path, "constraints.json"
            )

            if auto_create is True:
                self._auto_create()

            self._load_from_path()
        else:
            self._create_path()

    def _auto_create(self):
        """
        Check that ``vertices`` and ``edges`` directories exists, and if
        not create them and all the other required files and folders.
        """
        status = [
            os.path.exists(self.vertices_path),
            os.path.exists(self.edges_path),
        ]

        if not all(status):
            self._create_vertex_skel(self.path)
            self._create_edge_skel(self.path)

    def _load_from_path(self):
        """
        Scan through the given database path and load/import up all the
        relevant vertices, vertices constraints, and edges.
        """
        logging.info("Loading graph data from %r", self.path)
        self._load_vconstraints_from_path(self.vertices_constraints_path)
        self._load_vertices_from_path(self.vertices_path)
        self._load_edges_from_path(self.edges_path)
        logging.info("Completed %r graph import", self.path)

    def _load_vconstraints_from_path(self, path):
        """
        Open, parse and load the vertices constraints.

        :param path: Vertices constraints file to open, parse and import.
        :type path: :class:`str`
        """
        logging.info("Loading vertices constraints %r", path)
        with open(path) as vconstraints_fh:
            for each in json.load(vconstraints_fh):
                self.add_vertex_constraint(each["label"], each["key"])

    def _load_vertices_from_path(self, path):
        """
        Scan through the given path and load/import all the vertices.

        .. code::

            path
               |_ vertices
                    |_ constraints.json (file)
                    |_ labelA
                    |     |_ 0
                    |        |_ properties.json (file)
                    |
                    |_ labelB
                         |_ 1
                            |_ properties.json (file)

        :param path: Vertices Path to walk and import.
        :type path: :class:`str`
        """
        logging.info("Loading vertices from %r", path)
        sorted_to_import = sorted(
            _search_for_vertex_id(path),
            key=lambda x: x[0]
        )

        for ident, label, prop_file in sorted_to_import:
            properties = json.load(open(prop_file)) if prop_file else {}

            # reset the id to the id being loaded.
            self._id_tracker.vid = ident
            vertex = super(PersistentGraph, self).add_vertex(
                label, **properties
            )
            # due to pylint bug https://github.com/PyCQA/pylint/issues/379, we
            # need to disable assigning-non-slot errors
            vertex.path = os.path.join(path, label, str(ident))  # pylint: disable=assigning-non-slot

    def _load_edges_from_path(self, path):
        """
        Scan through the given path and load/import all the edges.

        .. code::

            path
               |_ edges
                     |_ label
                          |_0
                            |_ properties.json (file)
                            |_ head
                            |   |_ 0 -> ../../../vertices/0 (symlink)
                            |_ tail
                                |_ 1 -> ../../../vertices/1 (symlink)

        :param path: Edges Path to walk and import.
        :type path: :class:`str`
        :raises KeyError: If the head or tail of the edge being
            imported is unknown.
        """
        logging.info("Loading edges from %r", path)
        sorted_to_import = sorted(
            _search_for_edge_ids(path),
            key=lambda x: x[0]
        )

        for ident, head_id, label, tail_id, prop_file in sorted_to_import:
            properties = json.load(open(prop_file)) if prop_file else {}
            head = self.get_vertex(head_id)
            tail = self.get_vertex(tail_id)

            # reset the id to the id being loaded.
            self._id_tracker.eid = ident
            edge = super(PersistentGraph, self).add_edge(
                head,
                label,
                tail,
                **properties
            )

            # due to pylint bug https://github.com/PyCQA/pylint/issues/379, we
            # need to disable assigning-non-slot errors
            edge.path = os.path.join(path, label, str(ident))  # pylint: disable=assigning-non-slot

    def _create_vertex_skel(self, path):
        """
        Create a vertex skeleton path.

        :param path: Path to create the vertex skeleton structure in.
        :type path: :class:`str`
        """
        self.vertices_path = os.path.join(path, "vertices")
        os.makedirs(self.vertices_path)

        self.vertices_constraints_path = os.path.join(
            self.vertices_path, "constraints.json"
        )
        with open(self.vertices_constraints_path, "w") as constraint_fh:
            constraint_fh.write("[]")

    def _create_edge_skel(self, path):
        """
        Create a edge skeleton path.

        :param path: Path to create the edge skeleton structure in.
        :type path: :class:`str`
        """
        self.edges_path = os.path.join(path, "edges")
        os.makedirs(self.edges_path)

    def _create_path(self):
        """
        Create a complete database skeleton path.
        """
        self.path = mkdtemp(suffix="-ruruki-db")
        logging.info("Created temporary graph db path %r", self.path)
        self._create_vertex_skel(self.path)
        self._create_edge_skel(self.path)

    def add_vertex_constraint(self, label, key):
        super(PersistentGraph, self).add_vertex_constraint(label, key)
        with open(self.vertices_constraints_path, "w") as constraint_fh:
            data = []
            for label, key in self.get_vertex_constraints():
                data.append({"label": label, "key": key})
            json.dump(data, constraint_fh, indent=4)

    def add_vertex(self, label=None, **kwargs):
        vertex = super(PersistentGraph, self).add_vertex(label, **kwargs)
        # due to pylint bug https://github.com/PyCQA/pylint/issues/379, we
        # need to disable assigning-non-slot errors
        vertex.path = os.path.join(self.vertices_path, label, str(vertex.ident))  # pylint: disable=assigning-non-slot
        os.makedirs(vertex.path)
        os.makedirs(os.path.join(vertex.path, "in-edges"))
        os.makedirs(os.path.join(vertex.path, "out-edges"))

        with open(os.path.join(vertex.path, "properties.json"), "w") as fh:
            json.dump(vertex.properties, fh)

        return vertex

    def add_edge(self, head, label, tail, **kwargs):
        edge = super(PersistentGraph, self).add_edge(
            head, label, tail, **kwargs
        )

        # due to pylint bug https://github.com/PyCQA/pylint/issues/379, we
        # need to disable assigning-non-slot errors
        edge.path = os.path.join(self.edges_path, label, str(edge.ident))  # pylint: disable=assigning-non-slot
        head_path = os.path.join(edge.path, "head")
        tail_path = os.path.join(edge.path, "tail")

        os.makedirs(edge.path)
        os.makedirs(head_path)
        os.makedirs(tail_path)

        with open(os.path.join(edge.path, "properties.json"), "w") as fh:
            json.dump(edge.properties, fh)

        os.symlink(head.path, os.path.join(head_path, str(head.ident)))
        os.symlink(tail.path, os.path.join(tail_path, str(tail.ident)))

        os.symlink(
            edge.path,
            os.path.join(
                head.path,
                "out-edges",
                str(edge.ident)
            )
        )

        os.symlink(
            edge.path,
            os.path.join(
                tail.path,
                "in-edges",
                str(edge.ident)
            )
        )

        return edge

    def set_property(self, entity, **kwargs):
        super(PersistentGraph, self).set_property(entity, **kwargs)

        # Update the properties to the properties file
        properties_file = os.path.join(
            entity.path,
            "properties.json"
        )

        with open(properties_file, "w") as prop_file:
            json.dump(
                dict(
                    (k, v)
                    for k, v in entity.properties.items()
                    if k != "_path"
                ),
                prop_file,
                indent=4
            )

    def remove_edge(self, edge):
        super(PersistentGraph, self).remove_edge(edge)
        shutil.rmtree(edge.path)

    def remove_vertex(self, vertex):
        super(PersistentGraph, self).remove_vertex(vertex)
        shutil.rmtree(vertex.path)



#!/usr/bin/env python
import os.path
from setuptools import setup, find_packages


def auto_version_setup(**kwargs):
    pkg_name = kwargs["packages"][0]

    # Populate the "version" argument from the "VERSION" file.
    pkg_path = os.path.join(os.path.dirname(__file__), pkg_name)
    with open(os.path.join(pkg_path, "VERSION"), "r") as handle:
        pkg_version = handle.read().strip()
    kwargs["version"] = pkg_version

    # Make sure the "VERSION" file is included when we build the package.
    package_data = kwargs.setdefault("package_data", {})
    this_pkg_data = package_data.setdefault(pkg_name, [])
    if "VERSION" not in this_pkg_data:
        this_pkg_data.append("VERSION")

    setup(**kwargs)


auto_version_setup(
    name="ruruki",
    author="Jenda Mudron",
    author_email="Jenda.Mudron@optiver.com.au",
    maintainer="Jenda Mudron",
    maintainer_email="Jenda.Mudron@optiver.com.au",
    url="https://github.com/optiver/ruruki",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 2 :: Only",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Software Development :: Libraries",
    ],
    keywords=(
        "graph db momory database in-memory snail extraction tool "
        "POC proof concept temporary lightweight"
    ),
    description=(
        "Ruruki is a lightweight in-memory graph database which is ideal "
        "if you need a temporary graph database, or just doing a proof of "
        "concept and need a fast and reliable graph database."
    ),
    packages=find_packages(),
    package_data={
        "ruruki": [
            "VERSION",
        ],
    },
    install_requires=[],
)
